#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<stddef.h>
#include"texs.h"
#include"heap.h"
#include"library.h"

Public library *
library_create(long blocksize)

{
    library *lib;
    heap    *lib_heap;

    lib_heap = heap_create(blocksize);
    lib = (library *)heap_allocate(lib_heap, sizeof(library));
    lib->lib_on_which_heap = lib_heap;
    lib->root = NULL;
    lib->number_of_entries = 0;
    return(lib);
}

Public void
library_read(library *lib, FILE *libfile)

{
    static unsigned char string[256];

    while(read_string(libfile,  string, 256) >= 0)
       {
        next_line(libfile);
        if(*string == EOS)
   	    continue;
         if(StrCmp((char *)string, "end-of-file") == 0)
	     break;
        library_enter_entry(lib, string);
       }
}

Public void
write_tree(struct libnode *root, FILE *libfile)

{
  struct libnode *aktualny, 
	         *kolejny,
	         *q, 
	         *q1, 
	         *q2;

    if(root != NULL)
       {
	aktualny = root; 
	root->link = NULL;
	kolejny = NULL;
	while(aktualny != NULL)
	   {
	    q = aktualny; 
	    do
	       {
		fprintf(libfile, "%s\n", q->string);
		q1 = q->left; 
		q2 = q->right;
		if(q1 != NULL)
		   { 
		    q1->link = kolejny;
		    kolejny = q1;
		   }
		if(q2 != NULL)
		   { 
		    q2->link = kolejny;
		    kolejny = q2;
		   }
		q = q->link;
	       }while(q != NULL);
	    aktualny = NULL;
	    while(kolejny != NULL)
	       { 
		q = kolejny; 
		kolejny = q->link;
		q->link = aktualny; 
		aktualny = q;
	       }
	   }
       }
}

Public Logical
search_and_set(library *lib, struct libnode **dest, const unsigned char *string)

{
    register        int     comp;	
    register struct libnode *p2,
                            *p1;
    if(*dest == NULL)
       { 				/* brak klucza w drzewie; wstaw go */
	*dest = (struct libnode *)heap_allocate(lib->lib_on_which_heap,
			    sizeof(struct libnode) + strlen((char *)string));
	(*dest)->left = NULL;
	(*dest)->action_type = 0;
	(*dest)->right = NULL;
	strcpy((char *)(*dest)->string, (char *)string);
	lib->number_of_entries++;
	return(True);
       }
    if((comp = StrCmp((char *)string, (char *)(*dest)->string)) < 0)
       {
	if(search_and_set(lib, &((*dest)->left), string))
	   {					/*urosla lewa galaz*/
	    switch((*dest)->action_type)
	       {
		default:
		     fprintf(stderr, 
		     		"Internal error - illegal leaf change mode\n");
		     exit(-1);
		case 1:
		    (*dest)->action_type = 0;
		    return(False);
		case 0:
		    (*dest)->action_type = -1;
		    return(True);
		case -1:
		    p1 = (*dest)->left;
		    if(p1->action_type == -1)
		       {	 		/*pojedyncza zamiana LL*/
			(*dest)->left = p1->right;
			p1->right = (*dest);
			(*dest)->action_type = 0;
			*dest = p1;
		       }
		     else
		       { 			/*podwjna zamiana LP*/
			p2 = p1->right;
			p1->right = p2->left;
			p2->left = p1;
			(*dest)->left = p2->right;
			p2->right = *dest;
			if(p2->action_type == -1)
			    (*dest)->action_type = 1;
			 else
			    (*dest)->action_type = 0;
			if(p2->action_type == 1)
			    p1->action_type = -1;
			 else
			    p1->action_type = 0;
			*dest = p2;
		       }
		    (*dest)->action_type = 0;
		    return(False);
	        }
	    }
       }
     else if(comp > 0)
       {
	if(search_and_set(lib, &((*dest)->right), string))
	   { 					/*urosla prawa galaz */
	    switch((*dest)->action_type)
	       {
		default:
		     fprintf(stderr, 
		     		"Internal error - illegal leaf change mode\n");
		     exit(-1);
		case -1:	
		    (*dest)->action_type = 0;
		    return(False);
		case 0:
		    (*dest)->action_type = 1;
		    return(True);
		case 1: 			/*wywaz drzewo*/
		    p1 = (*dest)->right;
		    if(p1->action_type == 1)
		       { 			/*pojedyncza zamiana PP*/
			(*dest)->right = p1->left;
			p1->left = *dest;
			(*dest)->action_type = 0;
			(*dest) = p1;
		       }
		     else
		       { 			/*podwjna zamiana PL*/
			p2 = p1->left;
			p1->left = p2->right;
			p2->right = p1;
			(*dest)->right = p2->left;
			p2->left = (*dest);
			if(p2->action_type == 1)
			    (*dest)->action_type = -1;
			 else
			    (*dest)->action_type = 0;
			if(p2->action_type == -1)
			    p1->action_type = 1;
			 else
			    p1->action_type = 0;
			(*dest) = p2;
		       }
		    (*dest)->action_type = 0;
		    return(False);
	        }
	    }
       }
    return(False);
}

Public Logical
library_find_entry(library *lib, const unsigned char *string)

{
    register struct libnode *node;
    register int comp;

    if(lib->root == NULL)
        return(False);
    node = lib->root;
    while((comp = StrCmp((char *)node->string, (char *)string)) != 0)
       {
        if((comp > 0) && (node->left))
	   {
	    node = node->left;
	    continue;
	   }
         else if((comp < 0) && (node->right))
	   {
	    node = node->right;
	    continue;
	   }
         else
	   return(False);
       }
    return(True);
}
