#include <stdio.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>

#include "texs.h"
#include "heap.h"
#include "library.h"
#include "parse.h"

#ifdef MSC
#define setbuffer(file, buf, size)   setvbuf(file, buf, _IOFBF, size)
#endif

#ifdef BCC
#define setbuffer(file, buf, size)   setvbuf(file, buf, _IOLBF, size)
#endif

#ifdef  WCC
#define setbuffer(file, buf, size)   setvbuf(file, buf, _IOFBF, size)
#endif

#define IOBUF_LEN	8000

#define LIBRARY_BLOCK_LEN	4096 * 4

/* returnwerte:
-1: option error
-2: file error
-3: out of memory */

#define ENVPATHVAR "SPELL_PATH"
#define LANGVAR    "SPELL_LANG"

#ifdef MS_DOS
#define DIR_SEP  "\\"
#else
#define DIR_SEP  "/"
#endif

Local Logical  checking       = False, 
	       merging        = False;
Public Logical case_sensitive = False;
Public Logical first_char_checking  = True;
Public  Logical newlibcreate         = True;
Local  Logical library_statistics   = False;
Local  Logical silencium = False;
Local  Logical firstlib = True;

Local char *make_file_name(char *file_name, char *ext, Logical force);
Local void open_library(library *mainlib, char *file_name, Logical merging);
Local void open_info(char *file_name);
Local void print_help(void);
Local FILE *open_file(const char *file_name, const char *mode, 
							const char *message);

Local void signal_act(int);

Public library       *mainlib;
Local char *envpath;
Local char *spell_lang;
Public library       *newlib;

Public int 
main(int argc,  char *argv[])

{
    unsigned char *word;
    FILE          *texfile, 
                  *outfile, 
                  *infofile, 
                  *newlibfile;
            char  *newlibname   = NULL, 
		  *outputname   = NULL;
	    char  *libnamespace = NULL,
		  *texfilename  = NULL, 
		  *infofilename = NULL;

    envpath = getenv(ENVPATHVAR);
    spell_lang = getenv(LANGVAR);

#ifndef _WINDOWS
    signal(SIGINT, signal_act);
#endif

    if(argc < 3)
       {
        print_help();
        return(-1);
       }
    punctuation_check = True;
    mainlib = library_create(LIBRARY_BLOCK_LEN);
    outfile = stdout;

   {
    register int i;
    Logical inputfilefound = False;
    Logical optionfound = False ;

    for(i = 1; i < argc; i++)
       {
	if(argv[i][0] == '-')  /* wir haben eine option */
	   {
	    switch(argv[i][1])
	       {
	        case '?':        /* libhist schreiben */
		    print_help();
		    exit(0);
	        case 'h':        /* libhist schreiben */
		    library_statistics = True;
		    argv[i] = NULL;
		    break;
	        case 'i':
		    argv[i++] = NULL;
		    infofilename = make_file_name(argv[i], ".inf", False);
		    argv[i] = NULL;
		    break;
	        case 'l':
		    argv[i] = NULL;
		    newlibname = argv[++i];
		    argv[i] = NULL;
		    break;
	        case 'o':       /* name des outfiles, default *.out */
		    argv[i] = NULL;
		    outputname = argv[++i];
		    argv[i] = NULL;
		    break;
	        case 'C':       /* case sensitive search */
		    case_sensitive = True;
		    argv[i] = NULL;
		    break;
	        case 'm':       /* merge file with lib */
		    optionfound = True;
		    merging = True;
		    argv[i] = NULL;
		    break;
	        case 'c':       /* check file for spelling errors */
		    checking = True;
		    optionfound = True;
		    argv[i] = NULL;
		    break;
	        case 'p':
		    punctuation_check = False;
		    argv[i] = NULL;
		    break;
	        case 's':
		    silencium = True;
		    argv[i] = NULL;
		    break;
	        case 'n':       /* don't create  libfile */
		    argv[i] = NULL;
		    newlibcreate = False;
		    break;
	        case 'P':       
		    first_char_checking = False;
		    argv[i] = NULL;
		    break;
	        default:
		    fprintf(stderr, "invalid option! aborting...\n");
		    return(-1);
	       }
	   }
       }    
    if(!optionfound)
       {
	fprintf(stderr, "work mode undefined\n");
	return(-1);
       }
    if(merging && checking)
       {
        fprintf(stderr, "-c and -m are incompatible options !\n");
        return(-1);
       }
    init_parser_data();
    for(i = 1; i < argc; i++)
       {
        if(argv[i] != NULL)
	   {
	    if(inputfilefound)
		open_library(mainlib, argv[i], merging);
	     else
	       {
		char *tmpbuf;

		if(merging)
		    texfilename = make_file_name(argv[i], ".twl", False);
		 else
		    texfilename = make_file_name(argv[i], ".tex", False);
		inputfilefound = True;
		if(!silencium)
		    printf("input file: %s\n", texfilename);
		if((texfile = fopen(texfilename, "r")) == NULL)
		   {
		    fprintf(stderr, "could not open input file %s \n", 
		    						texfilename);
		    return(-2);
		   }
		tmpbuf = (char *)malloc(IOBUF_LEN * 4);
		if(tmpbuf != NULL)
		    setbuffer(texfile, tmpbuf, IOBUF_LEN * 4);
		if(newlibcreate)
		   {
		    if(newlibname == NULL && !merging)
			libnamespace = make_file_name(texfilename, 
								".twl", True);
		     else if(!merging)
			libnamespace = make_file_name(newlibname, 
								".twl", False);
		    if(!silencium && !merging)
			printf("new library: %s\n", libnamespace);
		   }
		if(spell_lang != NULL)
		    open_library(mainlib, spell_lang, merging);
	       }
	   }
       }
    if(!silencium)
        printf("\n");
    if(!inputfilefound)
       {
	fprintf(stderr, "input file name missing\n");
	return(-1);
       }
   }
    if(!merging)
       {
        char *tmpbuf;

        if(outputname)
	   {
	    if((outfile = fopen(outputname, "w")) == NULL)
	       {
	        fprintf(stderr, "could't open outputfile %s \n", outputname);
	        return(-2);
	       }
	    if(!silencium)
	        printf("output: %s\n", outputname);
	    tmpbuf = (char *)malloc(IOBUF_LEN * 4);
	    if(tmpbuf != NULL)
		setbuffer(outfile, tmpbuf, IOBUF_LEN * 4);
	   }
	if(spell_lang != NULL)
	    open_info(spell_lang);
	if(infofilename != NULL)
	    open_info(infofilename);
       }
    if(newlibcreate && !merging)
        newlib = library_create(LIBRARY_BLOCK_LEN);

    if(merging)
       {
        FILE *newlibfile;
        char *tmpbuf;

        if(!silencium)
 	    printf("merging input dictionary to %s\n", texfilename);
        if((newlibfile = fopen(texfilename, "r")) == NULL)
	   {
	    fprintf(stderr, "could not open word list file %s \n", texfilename);
	    return(-2);
	   }
	tmpbuf = (char *)malloc(IOBUF_LEN * 4);				/*WM */
	if(tmpbuf != NULL)					/*WM */
	    setbuffer(newlibfile, tmpbuf, IOBUF_LEN * 4);		/*WM */
        if(!silencium)
	   printf("reading: %s\n", texfilename);
        library_read(mainlib, newlibfile);
        fclose(newlibfile);

        if((newlibfile = fopen(texfilename, "w")) == NULL)
	   {
	    fprintf(stderr, "could't open otput library %s \n", texfilename);
	    return(-2);
	   }
	tmpbuf = (char *)malloc(IOBUF_LEN * 4);				/*WM */
	if(tmpbuf != NULL)					/*WM */
	   setbuffer(newlibfile, tmpbuf, IOBUF_LEN * 4);	/*WM */
        if(!silencium)
  	    printf("writing: %s\n", texfilename);
        library_write(mainlib, newlibfile);
        fclose(newlibfile);

        library_delete(mainlib);
        return(0);
       }

    if(!silencium)
	printf("start parsing...\n");

    init_parser(texfile, outfile, texfilename);

    parse(&catcodes[0]);

    library_delete(mainlib);

    if(outfile != stdout)
        fclose(outfile);
    fclose(texfile);
    if(newlibcreate)
       {
        char *tmpbuf;

        if((newlibfile = fopen(libnamespace, "w")) == NULL)
	   {
	    fprintf(stderr, "could't open output word list %s \n", 
	  							libnamespace);
	    return(-2);
	  }
       tmpbuf = (char *)malloc(IOBUF_LEN * 4);				/*WM */
       if(tmpbuf != NULL)					/*WM */
           setbuffer(newlibfile, tmpbuf, IOBUF_LEN * 4);	/*WM */
       if(!silencium)
           printf("writing new library %s\n", libnamespace);
       library_write(newlib, newlibfile);
       fclose(newlibfile);
       library_delete(newlib);
      }
    return(0);
}                                   /* end of main */

Local void 
print_help(void)

{
    printf("This is TeXspell, version 1.5, by J.Hannappel and E.Werner \n");
    printf("written with GNU C, 1991, additions ");
    printf("- Z.S.Kulesza, 1996 & W.Macewicz 1997\n\n");
    printf("TeXspell [<option>] <input file> [<dictionary file>]\n");
    printf("options:\n");
    printf(" -s\tsilencium mode\n");
    printf(" -c\tcheck textfile for errors\n");
    printf(" -m\tmerge word lists\n");
    printf("exactly one of the above options is required\n");
    printf(" -i\tnext parameter is the info file name\n");
    printf(" -l\tnext parameter is the output word list name; default: jobname.twl\n");
    printf(" -o\tnext parameter is the output file name; default: stdout\n");
    printf(" -C\tcase sensitive search; be sure your library is made for that!\n");
    printf(" -p\tsupress punctuation check\n");
    printf(" -P\tsupress check for capital letters at the beginnings of sentences\n");
    printf(" -n\tdon't create output word list\n");
    printf("\nFirst file name is TeX(t) file, the others are word lists\n");
    printf("Info and word list files may be placed in a directory specified\n");
    printf("by an environment variable SPELL_PATH\n");
    printf("\nfor additional help, check the manual\n");
}

Local FILE * 
open_file(const char *file_name, const char *mode, const char *message)

{
    FILE *tmp_file = NULL;
    char *fullname;
    char *tmpbuf;		/* WM */

    if((tmp_file = fopen(file_name, mode)) == NULL)
       {
	if(envpath != NULL)
	   {
	    fullname = (char *)malloc(strlen(envpath) + strlen(DIR_SEP) + 
							strlen(file_name) + 1);
	    if(fullname == NULL)
	       {
		fprintf(stderr, "memory overflow\n");
		exit(-3);
	       }
	    strcpy(fullname, envpath);
	    strcat(fullname, DIR_SEP);
	    strcat(fullname, file_name);
	    if((tmp_file = fopen(fullname, mode)) == NULL)
	       {
		fprintf(stderr, "could't open %s: %s \n", message, file_name);
		exit(-2);
	       }
	   }
	 else
	   {
	    fprintf(stderr, "Could't open %s: %s \n", message, file_name);
	    exit(-2);
	   }
       }
   tmpbuf = (char *)malloc(IOBUF_LEN);				/*WM */
   if(tmpbuf != NULL)					/*WM */
       setbuffer(tmp_file, tmpbuf, IOBUF_LEN);		/*WM */
    return(tmp_file);
}

Local void 
open_library(library *mainlib, char *file_name, Logical merging)

{
    FILE *libfile;
    char *libfilename;

    libfilename = make_file_name(file_name, ".twl", False);
    if(!silencium)
       {
	if(firstlib)
	    printf("open dictionary: %s", libfilename);
	 else
	    printf(", %s", libfilename);
       }
    libfile = open_file(libfilename, "r", "word list");
    firstlib = False;
    library_read(mainlib, libfile);
    fclose(libfile);
    if(file_name != libfilename)
        free(libfilename);
}

Local void 
open_info(char *file_name)

{
    FILE *libfile;
    char *libfilename;

    libfilename = make_file_name(file_name, ".inf", False);
    if(!silencium)
	printf("open info: %s\n", libfilename);
    libfile = open_file(libfilename, "r", "info file");
    read_parser_info(libfile);
    fclose(libfile);
    if(file_name != libfilename)
        free(libfilename);
}

Local char *
make_file_name(char *file_name, char *ext, Logical force)

{
    char *libfilename;

    if((libfilename = strchr(file_name, '.')) != NULL)
       {
	if(!force)
	   {
	    if(*(libfilename + 1) == EOS)
		*libfilename = EOS;
	    return(file_name);
	   }
       }
    if((libfilename = (char *)malloc(strlen(file_name) + 
						strlen(ext) + 1)) == NULL)
       {
	fprintf(stderr, "memory overflow\n");
	exit(-3);
       }
    strcpy(libfilename, file_name);
    if((file_name = strchr(libfilename, '.')) != NULL)
	*file_name = EOS;
    strcat(libfilename, ext);
    return(libfilename);
}

Local void 
signal_act(int x)

{
    fprintf(stderr, "\nAbort operation; CTRL-C\n");
    exit(0);
}
