#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include"texs.h"
#include"heap.h"
#include"library.h"
#include"parse.h"

#ifdef  DEBUG
Local enum parsemode modering[RINGSIZE];
Local int ringindex;
#define storemode(curr_mode_ptr) { \
                            ringindex++; \
			    ringindex %= RINGSIZE; \
			    modering[ringindex] = curr_mode_ptr; \
			    printf("%d ", modering[ringindex]);\
			}
#define preprevious_mode  (modering[ringindex > 0 ? ringindex - 1 :\
								 RINGSIZE - 1])
#define previous_mode     (modering[ringindex])
#define old_mode(age)     (modering[ringindex - age => 0 ? \
					ringindex-age :\
						RINGSIZE + ringindex - age])
#else
#define storemode(curr_mode_ptr)
#endif

#ifdef min
#undef min
#endif

#ifdef isdigit
#undef isdigit
#endif

#define min(a, b)          ((a) < (b) ? (a) : (b))
#define isdigit(c) 	  ((catcodes[c].type & IS_DIGIT) ? True : False)

Public volatile int column_number;
Public volatile int line_number;
Public Logical punctuation_check;
Local  Logical new_sentence;

Local void lowercase(unsigned char *string);
Local void chk_word(unsigned char *);
Local int line_offset;
Local struct active_definition *active_defs;
Local heap *active_heap;
Local enum parsemode mode_stack[MODESTACK];
Local enum parsemode *curr_mode_ptr;            /* pointer auf den mode_stack */
Local struct token  current_token;
Local FILE          *source;
Local FILE          *outputfile;
Local const   char  *texname;
Public struct token catcodes[256];
Local library *skipped_env,
              *skipped_com,
	      *special_com;
Local int activ_column_number;

Local struct token next_token(void);
Local void exit_parser(void);
Local void ignore_comment(void);
Local int execute_catcode(void);

Local int read_character(FILE *file, unsigned char *buffer);

Public void
init_parser_data(void)

{

#ifdef DEBUG
  for(ringindex = 0; ringindex < RINGSIZE; ringindex++)
      modering[ringindex] = no_mode;
  ringindex = 0;
#endif

  skipped_env = library_create(DEF_LIB_LEN);
  skipped_com = library_create(DEF_LIB_LEN);
  special_com = library_create(DEF_LIB_LEN);
  active_heap = heap_create(DEF_LIB_LEN);
  atexit(exit_parser);
  active_defs = NULL;

  { 					/* init catcodes */
    int c;

    for(c = 0; c <= 255; c++)
       {
	catcodes[c].catcode = catcode_illegal;
	catcodes[c].token = (unsigned char)c;
       }
    for(c = ' ' + 1; c < 128; c++)
	catcodes[c].catcode = catcode_other;

    for(c = '0'; c <= '9'; c++)
	catcodes[c].type |= IS_DIGIT;

    catcodes['.'].type |= IS_SENTENCE;
    catcodes['!'].type |= IS_SENTENCE;
    catcodes['?'].type |= IS_SENTENCE;
    catcodes['('].type |= IS_T_BEG;
    catcodes[')'].type |= IS_T_END;
    catcodes['.'].type |= IS_PUNC;
    catcodes[','].type |= IS_PUNC;
    catcodes[':'].type |= IS_PUNC;
    catcodes[';'].type |= IS_PUNC;
    catcodes['!'].type |= IS_PUNC;
    catcodes['?'].type |= IS_PUNC;

    catcodes['\f'].catcode = catcode_ignored;
    catcodes['\b'].catcode = catcode_ignored;
    catcodes['\a'].catcode = catcode_ignored;
    catcodes['\v'].catcode = catcode_ignored;
    catcodes['\r'].catcode = catcode_ignored;

    catcodes['\\'].catcode = catcode_escape;
    catcodes['{'].catcode = catcode_begingroup;
    catcodes['}'].catcode = catcode_endgroup;
    catcodes['$'].catcode = catcode_mathshift;
    catcodes['&'].catcode = catcode_alignmenttab;
    catcodes['\n'].catcode = catcode_endofline;
    catcodes['\n'].token = '\n';
    catcodes[' '].catcode = catcode_space;
    catcodes['\t'].catcode = catcode_space;
    catcodes['%'].catcode = catcode_comment;

    for(c = 'A'; c <= 'Z'; c++)
     {
      catcodes[c].catcode = catcode_letter;
      catcodes[c].upper = (unsigned char)c;
      catcodes[c].lower = (unsigned char)c - 'A' + 'a';
     }
    for(c = 'a'; c <= 'z'; c++)
     {
      catcodes[c].catcode = catcode_letter;
      catcodes[c].lower = (unsigned char)c;
      catcodes[c].upper = (unsigned char)c - 'a' + 'A';
     }
    catcodes[0].catcode = catcode_eof;
  } /* init catcodes */
}

Public void
read_parser_info(FILE *infofile)

{

    unsigned char key_word_value[MAXSTRING];
    unsigned char key_word[MAXSTRING];

    while((read_string(infofile, key_word, MAXSTRING) >= 0))
       {
	unsigned char *key_word_value_ptr;

	if(*key_word == EOS)
	   {
	    next_line(infofile);
	    continue;
	   }
        if(!StrCmp((char *)key_word, "end_of_file"))
  	   break;
        if(read_string(infofile, key_word_value, MAXSTRING) <= 0)
           {
	    fprintf(stderr, "Missing value of keyword %s\n", key_word);
	    exit(1);
	   }
	next_line(infofile);
        key_word_value_ptr = key_word_value;
        if(!StrCmp((char *)key_word, "illegal"))
  	   while(*key_word_value_ptr)
	       catcodes[*key_word_value_ptr++].catcode = catcode_illegal;
         else if(!StrCmp((char *)key_word, "escape"))
	   while(*key_word_value_ptr)
	       catcodes[*key_word_value_ptr++].catcode = catcode_escape;
	 else if(!StrCmp((char *)key_word, "begingroup"))
	    while(*key_word_value_ptr)
		catcodes[*key_word_value_ptr++].catcode = catcode_begingroup;
	 else   if(!StrCmp((char *)key_word, "endgroup"))
	    while(*key_word_value_ptr)
		catcodes[*key_word_value_ptr++].catcode = catcode_endgroup;
	 else  if(!StrCmp((char *)key_word, "mathshift"))
	    while(*key_word_value_ptr)
		catcodes[*key_word_value_ptr++].catcode = catcode_mathshift;
	 else  if(!StrCmp((char *)key_word, "alignmenttab"))
	    while(*key_word_value_ptr)
		catcodes[*key_word_value_ptr++].catcode = catcode_alignmenttab;
	 else  if(!StrCmp((char *)key_word, "endofline"))
	    while(*key_word_value_ptr)
		catcodes[*key_word_value_ptr++].catcode = catcode_endofline;
	 else   if(!StrCmp((char *)key_word, "space"))
	    while(*key_word_value_ptr)
		catcodes[*key_word_value_ptr++].catcode = catcode_space;
	 else  if(!StrCmp((char *)key_word, "letter"))
	    while(*key_word_value_ptr)
		catcodes[*key_word_value_ptr++].catcode = catcode_letter;
         else  if(!StrCmp((char *)key_word, "active"))
  	   {
  	    register int               i,
  	                               num_of_lines;
  	    struct   active_definition *ad;
  	    unsigned char              expansion[MAXSTRING];

  	    catcodes[*key_word_value_ptr++].catcode = catcode_active;
  	    num_of_lines = atoi((char *)key_word_value_ptr);
  	    ad = (struct active_definition *)heap_allocate(active_heap,
  	  			sizeof(struct active_definition) +
  			sizeof(struct active_expansion) * (num_of_lines - 1));
  	    ad->active_char = *key_word_value;
  	    ad->num_of_lines = num_of_lines;
  	    ad->next = active_defs;
  	    active_defs = ad;
  	
  	    for(i = 0; i < num_of_lines; i++)
  	       {
  	        if((read_character(infofile,
  		    &(((ad->expansion)[i]).active_char_parameter)) == EOF) ||
  			  (read_string(infofile, expansion, MAXSTRING) <= 0))
  	           {
  		    fprintf(stderr, "end of file!\n");
  		    exit(1);
  		   }
  		next_line(infofile);
  	        ((ad->expansion)[i]).active_expansion =
  				    (unsigned char *)heap_allocate(active_heap,
  						strlen((char *)expansion) + 1);
  	        strcpy((char *)((ad->expansion)[i]).active_expansion,
  							    (char *)expansion);
  	       }
  	  }
        else  if(!StrCmp((char *)key_word, "other"))
  	   while(*key_word_value_ptr)
  	       catcodes[*key_word_value_ptr++].catcode = catcode_other;
        else  if(!StrCmp((char *)key_word, "ignored"))
  	   while(*key_word_value_ptr)
  	       catcodes[*key_word_value_ptr++].catcode = catcode_ignored;
        else  if(!StrCmp((char *)key_word, "comment"))
  	   while(*key_word_value_ptr)
  	       catcodes[*key_word_value_ptr++].catcode = catcode_comment;
        else  if(!StrCmp((char *)key_word, "skipped_command"))
  	   library_enter_entry(skipped_com, key_word_value + 1);
        else  if(!StrCmp((char *)key_word, "punctuation"))
  	   while(*key_word_value_ptr)
  	       catcodes[*key_word_value_ptr++].type |= IS_PUNC;
        else  if(!StrCmp((char *)key_word, "sentenceend"))
  	   while(*key_word_value_ptr)
  	       catcodes[*key_word_value_ptr++].type |= IS_SENTENCE;
        else  if(!StrCmp((char *)key_word, "text_begingroup"))
  	   while(*key_word_value_ptr)
  	       catcodes[*key_word_value_ptr++].type |= IS_T_BEG;
        else  if(!StrCmp((char *)key_word, "text_endgroup"))
  	   while(*key_word_value_ptr)
  	       catcodes[*key_word_value_ptr++].type |= IS_T_END;
        else  if(!StrCmp((char *)key_word, "skipped_environment"))
  	   library_enter_entry(skipped_env, key_word_value);
        else  if(!StrCmp((char *)key_word, "special_command"))
  	   library_enter_entry(special_com, &key_word_value[1]);
        else  if(!StrCmp((char *)key_word, "capsmall"))
          {
  	   catcodes[key_word_value[0]].upper = key_word_value[0];
  	   catcodes[key_word_value[0]].lower = key_word_value[1];
  	   catcodes[key_word_value[1]].upper = key_word_value[0];
  	   catcodes[key_word_value[1]].lower = key_word_value[1];
          }
        else if(!StrCmp((char *)key_word, "end_of_file"))
  	   break;
        else
  	  {
  	   fprintf(stderr,
  		"\nInvalid command %s ignored; check infofile\n\b", key_word);
  	   exit(1);
  	  }
       }
}

Public void
init_parser(FILE *sourcefile, FILE *outfile, const char *sourcename)

{
    outputfile = outfile;
    source = sourcefile;
    texname = sourcename;
    line_number = 1;
    activ_column_number = 0;
    line_offset = 0;

    current_token = next_token();
    curr_mode_ptr = mode_stack;
    *(curr_mode_ptr++) = normal_mode;
    *curr_mode_ptr = normal_mode;
    new_sentence = True;
#ifdef LIB_DEBUG
    library_write(skipped_env, stdout);
    library_write(skipped_com, stdout);
    library_write(special_com, stdout);
#endif
}


Local unsigned char command_buffer[MAXSTRING];
Local unsigned char *command_buf_ptr;
Local unsigned char env_name_buffer[MAXSTRING];
Local unsigned char *env_name_buffer_ptr;
Local unsigned char endbuffer[MAXSTRING];
Local unsigned char *end_buf_ptr;
Local unsigned char *word_ptr;
Local          int word_len = 0;
Local unsigned char wordbuffer[MAX_WORD_LEN + 1];


Public void
parse(struct token *end_token)

{
    int ret_exec;

    command_buf_ptr = command_buffer;
    word_ptr = wordbuffer;
    word_len = 0;
    line_number += line_offset;
    line_offset = 0;

    while(True)
       {
        if(word_ptr < wordbuffer)
  	   {
 	    word_ptr = wordbuffer;
 	    fprintf(outputfile, "%d %d: internall error\n", 
	    					line_number, column_number);
 	   }
        if(word_len > MAX_WORD_LEN)
 	   {
 	    fprintf(outputfile, "%d %d: word to long\n",
						line_number, column_number);
	    if(word_len > 0)
		chk_word(wordbuffer);
 	    continue;
 	   }
        if(curr_mode_ptr <= mode_stack)
 	   {
 	    line_number += line_offset;
 	    line_offset = 0;
 	    fprintf(outputfile, "%d %d: stack underflow\n",
						line_number, column_number);
 	    curr_mode_ptr = mode_stack + 1;	
 	    *curr_mode_ptr = normal_mode;
 	   }
        if(curr_mode_ptr >= &mode_stack[ MODESTACK - 1]) /* stack overflow */
 	   {
 	    line_number += line_offset;
 	    line_offset = 0;
 	    fprintf(outputfile, "%d %d: stack overflow\n",
						   line_number, column_number);
 	    curr_mode_ptr = mode_stack + 1;	
 	    *curr_mode_ptr = normal_mode;
 	   }
 	switch(*curr_mode_ptr)
 	   {
	    case normal_mode:
	        if(current_token.catcode == catcode_letter)
		   {
	  	    column_number = activ_column_number;
		    while((current_token.catcode == catcode_letter) ||
				    (current_token.catcode == catcode_comment))
		       {
			if(current_token.catcode == catcode_comment)
			    ignore_comment();
		        if(current_token.catcode == catcode_letter)
			   {
			    *word_ptr++ = current_token.token;
			    word_len++;
			    current_token = next_token();
			   }
		       }
		   if(punctuation_check && (current_token.type & IS_T_BEG))
		      {
		       fprintf(outputfile,
				    "%d %d: missing space before %c\n",
					    line_number, activ_column_number,
							current_token.token);
		       }
		   }
		if((ret_exec = execute_catcode()) == WORD_MODE)
		   {
		    if(word_len > 0)
			chk_word(wordbuffer);
		   }
                 else if(ret_exec == EOF_MODE)
		    {
		    *word_ptr = EOS;
		    if(word_len > 0)
			chk_word(wordbuffer);
		     return;
		    }
	        break;
	    case escape_mode:     			/* escape_mode */
	        switch(current_token.catcode)
		   {
		    case catcode_endofline:
		    case catcode_space:
		        *word_ptr = EOS;
		        storemode(*curr_mode_ptr);
		        *curr_mode_ptr = space_mode;
		        if(word_len > 0)
		            chk_word(wordbuffer);
		        break;
		    case catcode_escape:
		        *word_ptr = EOS;    /* \\ = \n word_ptr eins runter,
						um alten backslash zu tilgen  */
		        current_token = next_token();
		        storemode(*curr_mode_ptr);
		        curr_mode_ptr--;
		        if(word_len > 0)
			    chk_word(wordbuffer);
		        break;
		    default:			/* \<non letter> */
			*command_buf_ptr++ = current_token.token;
			current_token = next_token();
			*command_buf_ptr = 0;
			if(library_find_entry(skipped_com, command_buffer))
			   {
			    storemode(*curr_mode_ptr);
			    *curr_mode_ptr = extra_space_mode;
			   }
			 else
			   {
			    storemode(*curr_mode_ptr);
			    curr_mode_ptr--;
			    *word_ptr = EOS;
			    if(word_len > 0)
				chk_word(wordbuffer);
			   }
			break;
		    case catcode_letter:
		        command_buf_ptr = command_buffer;
		        do
			   {
			    *command_buf_ptr++ = current_token.token;
			    current_token = next_token();
		           }while(current_token.catcode == catcode_letter);
		        *command_buf_ptr = 0;
			while((current_token.catcode == catcode_space) ||
				(current_token.catcode == catcode_endofline))
			    current_token = next_token();
			if(current_token.catcode == catcode_comment)
			    ignore_comment();
		        if(!StrCmp((char *)command_buffer, "begin") &&
				    current_token.catcode == catcode_begingroup)
			   {
			    storemode(*curr_mode_ptr);
			    env_name_buffer_ptr = env_name_buffer;
			    *env_name_buffer_ptr = EOS;
			    current_token = next_token();
			    while((current_token.catcode != catcode_endgroup) &&
				(current_token.catcode != catcode_endofline) &&
				(current_token.catcode != end_token->catcode))
			       {
				if(current_token.catcode == catcode_comment)
				    ignore_comment();
			        *env_name_buffer_ptr++ = current_token.token;
				current_token = next_token();
			       }
			    *env_name_buffer_ptr = EOS;
			    if(current_token.catcode == catcode_endgroup)
				current_token = next_token();
			    *curr_mode_ptr = env_mode;
			    break;
			   }
		        if(!StrCmp((char *)command_buffer, "end") &&
				    current_token.catcode == catcode_begingroup)
			   {
			    storemode(*curr_mode_ptr);
			    while((current_token.catcode != catcode_endgroup) &&
				(current_token.catcode != catcode_endofline) &&
				(current_token.catcode != end_token->catcode))
			       {
				if(current_token.catcode == catcode_comment)
				    ignore_comment();
			        *env_name_buffer_ptr++ = current_token.token;
				current_token = next_token();
			       }
			    *env_name_buffer_ptr = EOS;
			    if(current_token.catcode == catcode_endgroup)
				current_token = next_token();
			    *curr_mode_ptr = parameters_mode;
			    break;
			   }
		        if(library_find_entry(special_com, command_buffer))
			   {
			    *word_ptr = '\\';
			    strcpy((char *)(word_ptr + 1),
							(char *)command_buffer);
			    word_ptr += (command_buf_ptr - command_buffer) + 1;
			    *word_ptr++ = '@';
			    word_len += (command_buf_ptr - command_buffer) + 1;
			    storemode(*curr_mode_ptr);
			    *curr_mode_ptr = extra_space_mode;
			   }
		         else if(library_find_entry(skipped_com,
			 					command_buffer))
		  	   {
			    storemode(*curr_mode_ptr);
			    *curr_mode_ptr = extra_space_mode;
			   }
			 else
			   {
			    *word_ptr = EOS;
			    storemode(*curr_mode_ptr);
			    *curr_mode_ptr = parameters_mode;
			    new_sentence = False;
			    if(word_len > 0)
			        chk_word(wordbuffer);
			   }
		        break;
		    case catcode_ignored:
		        ;
		        break;
		    case catcode_eof:
			*word_ptr = EOS;
			if(word_len > 0)
			    chk_word(wordbuffer);
		        return;
 	
		    }           /* local switch */
	         break;
	    case math_mode:
		do
		   {
		    if(current_token.catcode == end_token->catcode)
		       {
			*word_ptr = EOS;
			if(word_len > 0)
			    chk_word(wordbuffer);
			return;
		       }
		    if(current_token.catcode == catcode_comment)
			ignore_comment();
		    current_token = next_token();
		   }while(current_token.catcode != catcode_mathshift);
		line_number += line_offset;
		line_offset = 0;
		current_token = next_token();
	        --curr_mode_ptr;
		break;
	    case display_mode:
		while(True)
		   {
		    if(current_token.catcode == end_token->catcode)
		       {
			*word_ptr = EOS;
			if(word_len > 0)
			    chk_word(wordbuffer);
			return;
		       }
		    if(current_token.catcode == catcode_comment)
			ignore_comment();
		    if(current_token.catcode == catcode_mathshift)
		       {
			current_token = next_token();
			if(current_token.catcode == catcode_mathshift)
			   {
			    --curr_mode_ptr;
			    current_token = next_token();
			    break;
			   }
			 else
			   {
			    storemode(*curr_mode_ptr);
			    *++curr_mode_ptr = math_mode;
			    break;
			   }
		       }
		    current_token = next_token();
		   }
		line_number += line_offset;
		line_offset = 0;
		break;
	    case env_mode:
		if(library_find_entry(skipped_env, env_name_buffer))
		   {
		    storemode(*curr_mode_ptr);
		    *curr_mode_ptr = skip_mode;
		    break;
		   }
		storemode(*curr_mode_ptr);
		*curr_mode_ptr = parameters_mode;
		if(word_len > 0)
		    chk_word(wordbuffer);
		break;
 	
	    case other_mode:
	      if(punctuation_check)
		{
		  struct token mishandled_token = current_token;

		  current_token = next_token();
		  if(current_token.catcode == end_token->catcode)
		     {
		      *word_ptr = EOS;
		      if(word_len > 0)
			  chk_word(wordbuffer);
		      return;
		     }
		  if(current_token.catcode == catcode_comment)
		      ignore_comment();
		  printf("t=%c\n", mishandled_token.token);
		  if(!isdigit(mishandled_token.token) &&
					(current_token.type & IS_T_BEG))
		     {
		       fprintf(outputfile,
				    "%d %d: missing space before %c\n",
					    line_number, activ_column_number,
							current_token.token);
		     }
		
		  if(mishandled_token.type & IS_PUNC)
		     {
		      if((mishandled_token.type & IS_SENTENCE) &&
					   !isdigit(current_token.token) &&
				   (current_token.catcode != catcode_escape) &&
				  (current_token.catcode != catcode_endgroup))
			   new_sentence = True;
		      if((current_token.catcode != catcode_space) &&
			    (current_token.catcode != catcode_endofline) &&
					   !isdigit(current_token.token) &&
				   (current_token.catcode != catcode_escape) &&
			       (current_token.catcode != catcode_endgroup) &&
			    (current_token.catcode != catcode_alignmenttab) &&
				  !(current_token.type & IS_T_END) &&
				    (current_token.catcode != catcode_other))
			 {		
			  fprintf(outputfile,
					"%d %d: missing space after %c\n",
					  line_number, activ_column_number,
							mishandled_token.token);
			 }
		     }
		   else if(mishandled_token.type & IS_T_BEG)
		     {
		      if((current_token.catcode == catcode_space) ||
			    (current_token.catcode == catcode_endofline) ||
			    (current_token.catcode == catcode_alignmenttab))
			 {		
			  fprintf(outputfile,
					"%d %d: extra space after %c\n",
					  line_number, activ_column_number,
						    mishandled_token.token);
			 }
		     }
		   else if(mishandled_token.type & IS_T_END)
		     {
		      if((current_token.catcode != catcode_space) &&
			    (current_token.catcode != catcode_endofline) &&
			    (current_token.catcode != catcode_alignmenttab) &&
			    (current_token.catcode != catcode_endgroup) &&
				  !(current_token.type & IS_T_END) &&
				    (current_token.catcode != catcode_other))
			 {		
			  fprintf(outputfile,
					"%d %d: missing space after %c\n",
					      line_number, activ_column_number,
							mishandled_token.token);
			 }
		     }
		}
	      else
		  current_token = next_token(); /* if punctuation_check */
	      storemode(*curr_mode_ptr);
	      curr_mode_ptr--;
	      break;
 	
	    case space_mode:
	    case extra_space_mode:
	      while((current_token.catcode == catcode_space) ||
				(current_token.catcode == catcode_endofline))
		current_token = next_token();
	      if(current_token.catcode == catcode_comment)
		  ignore_comment();
	      if(current_token.catcode == end_token->catcode)
	         {
		  *word_ptr = EOS;
		  chk_word(wordbuffer);
		  return;
	         }
	      line_number += line_offset;
	      line_offset = 0;
	      if(punctuation_check)
		  if((current_token.type & IS_PUNC) ||
					    (current_token.type & IS_T_END))
		      if(*curr_mode_ptr != extra_space_mode)
			  fprintf(outputfile,
				    "%d %d: extra space before %c\n",
					    line_number, activ_column_number,
							current_token.token);
	      storemode(*curr_mode_ptr);
	      curr_mode_ptr--;
	      break;
	    case skip_mode:        /* environments{env_name_buffer}
							    werden geskipt    */
	      do {
		do {
		  do {
		    do {
		      do {
			if(current_token.catcode == end_token->catcode)
			    return;
			while(current_token.catcode != catcode_escape)
			    current_token = next_token();
			if(current_token.catcode == end_token->catcode)
			    return;
			current_token = next_token();
		      }while(current_token.token != 'e');
		      current_token = next_token();
		    }while(current_token.token != 'n');
		    current_token = next_token();
		  }while(current_token.token != 'd');
		  current_token = next_token();
		}while(current_token.catcode != catcode_begingroup);
		current_token = next_token();
		end_buf_ptr = endbuffer;
		while(current_token.catcode == catcode_letter)
		   {
		    *end_buf_ptr++ = current_token.token;
		    current_token = next_token();
		   }
		*end_buf_ptr = 0;
	      }while(StrCmp((char *)endbuffer, (char *)env_name_buffer) != 0);
	      current_token = next_token();
	      if(current_token.catcode == end_token->catcode)
		 {
		  *word_ptr = EOS;
		  chk_word(wordbuffer);
		  return;
		 }
	      storemode(*curr_mode_ptr);
	      curr_mode_ptr--;
	      break;
	    case comment_mode:
	      ignore_comment();
	      if(current_token.catcode == end_token->catcode)
		 {
		  *word_ptr = EOS;
		  chk_word(wordbuffer);
		  return;
		 }
	      storemode(*curr_mode_ptr);
	      curr_mode_ptr--;
	      break;
	    case active_mode:
	      {
		unsigned char wrong_token;

		struct active_definition *def;
		register int i = 0;
 	
		storemode(*curr_mode_ptr);
		curr_mode_ptr--;
		def = active_defs;
		while(def->active_char != current_token.token)
		    def = def->next;
		current_token = next_token();
		while((((def->expansion)[i]).active_char_parameter !=
			    current_token.token) && (i < def->num_of_lines))
		    i++;
		wrong_token = current_token.token;
		current_token = next_token();
		if(i == def->num_of_lines)
		   {
		    fprintf(outputfile,
			"%d %d: active  %c with unknown parameter %c\n",
					     line_number, activ_column_number,
					    def->active_char, wrong_token);
		   }
		 else
		   {
		    if(word_len +
		         strlen((char *)((def->expansion)[i]).active_expansion)
								>= MAX_WORD_LEN)
	               {
			fprintf(outputfile, "%d %d: word to long\n",
						line_number, column_number);
			*word_ptr = EOS;
			chk_word(wordbuffer);
			break;
		       }
		    strcpy((char *)word_ptr,
			    (char *)((def->expansion)[i]).active_expansion);
		    word_ptr +=
			strlen((char *)((def->expansion)[i]).active_expansion);
		    word_len +=
			strlen((char *)((def->expansion)[i]).active_expansion);
		   }
	      }
	      break;
	    case in_optional_parameters_mode:
	        if(current_token.token == ']')
		   {
		    *word_ptr = EOS;
		    current_token = next_token();
		    if(current_token.catcode == catcode_comment)
			ignore_comment();
		    if(current_token.token == '*')
			current_token = next_token();
		    if(current_token.catcode == catcode_comment)
			ignore_comment();
		    if(current_token.catcode == catcode_begingroup)
		       {
			*curr_mode_ptr = in_default_parameters_mode;
			current_token = next_token();
			if(word_len > 0)
			    chk_word(wordbuffer);
		       }
		     else if(current_token.token == '[')
		       {
			current_token = next_token();
			if(word_len > 0)
			    chk_word(wordbuffer);
		       }
		     else
		       {
		        curr_mode_ptr--;
			if(word_len > 0)
			    chk_word(wordbuffer);
		       }
		   }
	         else
		   {
		    if(current_token.catcode == catcode_letter)
		       {
			column_number = activ_column_number;
			while((current_token.catcode == catcode_letter) ||
				(current_token.catcode == catcode_comment))
			   {
			    if(current_token.catcode == catcode_comment)
				ignore_comment();
			    if(current_token.catcode == catcode_letter)
			       {
				*word_ptr++ = current_token.token;
				word_len++;
				current_token = next_token();
			       }
			   }
		       if(punctuation_check && (current_token.type & IS_T_BEG))
			  {
			   fprintf(outputfile,
					"%d %d: missing space before %c\n",
					    line_number, activ_column_number,
						    current_token.token);
			   }
		       }
		    if((ret_exec = execute_catcode()) == WORD_MODE)
			 chk_word(wordbuffer);
		     else if(ret_exec == EOF_MODE)
		        {
		         *word_ptr = EOS;
			 chk_word(wordbuffer);
			 return;
			}
		    break;
		   }
	        break;
	    case in_default_parameters_mode:
		if(current_token.catcode == catcode_endgroup)
		   {
		    *word_ptr = EOS;
		    current_token = next_token();
		    if(current_token.catcode == catcode_comment)
			ignore_comment();
		    if(current_token.token == '*')
			current_token = next_token();
		    if(current_token.catcode == catcode_comment)
			ignore_comment();
		    if(current_token.token == '[')
		       {
			*curr_mode_ptr = in_optional_parameters_mode;
			current_token = next_token();
			if(word_len > 0)
			    chk_word(wordbuffer);
		       }
		     else if(current_token.catcode == catcode_begingroup)
		       {
			current_token = next_token();
			if(word_len > 0)
			    chk_word(wordbuffer);
		       }
		     else
		       {
		        curr_mode_ptr--;
			if(word_len > 0)
			    chk_word(wordbuffer);
		       }
		   }
	         else
		   {
		    if(current_token.catcode == catcode_letter)
		       {
			column_number = activ_column_number;
			while((current_token.catcode == catcode_letter) ||
				(current_token.catcode == catcode_comment))
			   {
			    if(current_token.catcode == catcode_comment)
				ignore_comment();
			    if(current_token.catcode == catcode_letter)
			       {
				*word_ptr++ = current_token.token;
				word_len++;
				current_token = next_token();
			       }
			   }
		       if(punctuation_check && (current_token.type & IS_T_BEG))
			  {
			   fprintf(outputfile,
					"%d %d: missing space before %c\n",
					    line_number, activ_column_number,
						    current_token.token);
			   }
			if(current_token.catcode == catcode_endgroup)
			    break;
		       }
		    if((ret_exec = execute_catcode()) == WORD_MODE)
			chk_word(wordbuffer);
		     else if(ret_exec == EOF_MODE)
		       {
		         *word_ptr = EOS;
			 chk_word(wordbuffer);
			 return;
		       }
		    break;
		   }
	        break;
	    case parameters_mode:
		while((current_token.catcode == catcode_space) ||
			(current_token.catcode == catcode_endofline))
		    current_token = next_token();
		if(current_token.catcode == catcode_comment)
		    ignore_comment();
		if(current_token.token == '*')
		    current_token = next_token();
	        if(current_token.token == '[')
		   {
		    *curr_mode_ptr = in_optional_parameters_mode;
		    current_token = next_token();
		    if(current_token.catcode == catcode_comment)
			ignore_comment();
		   }
		 else if(current_token.catcode == catcode_begingroup)
		   {
		    *curr_mode_ptr = in_default_parameters_mode;
		    current_token = next_token();
		    if(current_token.catcode == catcode_comment)
			ignore_comment();
		   }
	         else
		    *curr_mode_ptr = extra_space_mode;
                break;
	    default:
	        break;
 	    }   /* switch */
       }      /* while 1 */
}          /* parse */

Local void
exit_parser(void)

{
  library_delete(skipped_com);
  library_delete(skipped_env);
  library_delete(special_com);
  heap_delete(active_heap);
}

Local void
ignore_comment(void)

{
      while(True)
         {
	  while(current_token.catcode != catcode_endofline)
	     {
	      current_token = next_token();
	      if(current_token.catcode == catcode_eof)
		  return;
	     }
	  do
	     {
	      current_token = next_token();
	     }while(current_token.catcode == catcode_space);
	  if(current_token.catcode != catcode_comment)
	      return;
       }
}

Local int
execute_catcode(void)

{
    switch(current_token.catcode)
       {
	case catcode_escape:
	    storemode(*curr_mode_ptr);
	    command_buf_ptr = command_buffer;
	    *++curr_mode_ptr = escape_mode;
	    current_token = next_token();
	    break;
	case catcode_begingroup:
	    storemode(*curr_mode_ptr);
	    *++curr_mode_ptr = normal_mode;
	    current_token = next_token();
	    break;
	case catcode_endgroup:
	    storemode(*curr_mode_ptr);
	    --curr_mode_ptr;
	    current_token = next_token();
	    break;
	case catcode_mathshift:
	    storemode(*curr_mode_ptr);
	    *++curr_mode_ptr = math_mode;
	    current_token = next_token();
	    if(current_token.catcode == catcode_mathshift)
	       {
		*curr_mode_ptr = display_mode;
		current_token = next_token();
	       }
	    *word_ptr = EOS;
	    if(word_len > 0)
		return(WORD_MODE);
	    break;
	case catcode_endofline:
	case catcode_space:
	    storemode(*curr_mode_ptr);
	    *++curr_mode_ptr = space_mode;
	    *word_ptr = EOS;
	    if(word_len > 0)
		return(WORD_MODE);
	    break;
	case catcode_alignmenttab:
	case catcode_other:
	    storemode(*curr_mode_ptr);
	    *++curr_mode_ptr = other_mode;
	    *word_ptr = EOS;
	    if(word_len > 0)
		return(WORD_MODE);
	    break;
	case catcode_active:
	    storemode(*curr_mode_ptr);
	    *++curr_mode_ptr = active_mode;
	    break;
	case catcode_comment:
	    storemode(*curr_mode_ptr);
	    *++curr_mode_ptr = comment_mode;
	    break;
	case catcode_ignored:
	    ;
	    break;
	case catcode_eof:
	    return(EOF_MODE);
	}
    return(NORMAL_MODE);
}

Public int
read_string(FILE *file, unsigned char *buffer, int max_len)

{
    register int c,
                 num = 0;

    do
      {
       c = fgetc(file);
       if(c == ' ' || c == '\t' || c == '\v' || c == '\f')
	    continue;
        else
	    break;
      }while(c != EOF);
    if(c == EOF)
       {
	*buffer = EOS;
	return(-1);
       }
    if(c == '#')
       {
	while((c = fgetc(file)) != EOF)
	   {
	    if(c == '\n')
	       {
		ungetc(c, file);
		*buffer = EOS;
	        return(0);
	       }
	   }
	*buffer = EOS;
	return(-1);
       }
    max_len--;
    while(max_len > 0)
       {
        if(c == ' ' || c == '\t' || c == '\v' || c == '\f' || c == '\n')
	    break;
	*buffer++ = (unsigned char)c;
	num++;
        max_len--;
        c = fgetc(file);
        if(c == EOF)
	   {
	    *buffer = EOS;
	    return(-1);
	   }
       }
    if(c == '\n')
        ungetc(c, file);
    *buffer = EOS;
    return(num);
}

Public void
next_line(FILE *file)

{
    register int c;

    while((c = fgetc(file)) != EOF)
       {
        if(c == '\n')
	   break;
       }
}

Local int
read_character(FILE *file, unsigned char *buffer)

{
    register int c;

    do
      {
       c = fgetc(file);
       if(c == ' ' || c == '\t' || c == '\v' || c == '\f')
	    continue;
        else
	    break;
      }while(c != EOF);
    if(c == EOF)
       {
	*buffer = EOS;
	return(EOF);
       }
    *buffer = (unsigned char)c;
    return(!(EOF));
}

Local void
chk_word(unsigned char * word)

{
    if(first_char_checking)
	if(new_sentence)
	   {
	    if(islower(*word))
	      fprintf(outputfile, 
	    "%d %d: capital letter expected after presumed end of sentence\n", 
						   line_number, column_number);
	   }
    new_sentence = False;
    line_number += line_offset;
    line_offset = 0;
    word_ptr = wordbuffer;
    if(word_len > MIN_WORD_LEN)
       {
	if(!library_find_entry(mainlib, word))
	   {
	    if(!case_sensitive)
	       {
		lowercase(word);
		if(library_find_entry(mainlib, word))
		   {
		    word_len = 0;
		    return;
		   }
	       }
	    fprintf(outputfile, "%d %d: unknown %s\n",
					    line_number, column_number, word);
	    if(newlibcreate) 
	       library_enter_entry(newlib, word);
	   }
       }
    word_len = 0;
}

Local void 
lowercase(unsigned char *string)

{

    while(*string = Tolower(*string))
        string++;
}                                   /* lowercase() */

Local struct token
next_token(void)

{
    register unsigned char curr_char;

    while(True)
       {
        curr_char = (unsigned char)getc(source);
        activ_column_number++;
        if(feof(source))
	    break;
        if(curr_char == '\n')
           {
	    activ_column_number = 0;
	    line_offset++;
           }
        if(catcodes[curr_char].catcode == catcode_illegal)
 	    fprintf(outputfile,
	         "%d %d: illegal character %02x\n", line_number,
						activ_column_number, curr_char);
         else if(catcodes[curr_char].catcode != catcode_ignored)
 	    return(catcodes[curr_char]);
       }
    return(catcodes[0]);
} /* next_token */
