enum catcode {
  catcode_escape, 
  catcode_begingroup, 
  catcode_endgroup, 
  catcode_mathshift, 
  catcode_alignmenttab, 
  catcode_endofline, 
  catcode_space, 
  catcode_letter, 
  catcode_other, 
  catcode_active, 
  catcode_comment, 
  catcode_ignored, 
  catcode_eof,
  catcode_illegal
};

struct token 
   {
    enum     catcode catcode;
    unsigned char    token;
    unsigned char    lower;
    unsigned char    upper;
    unsigned char    type;
   };

enum parsemode
   {
    no_mode, 		
    normal_mode, 	/* normale buchstaben                       */
    escape_mode, 	/* befehle                                  */
    env_mode, 		/* untersuchung von environment-parametern  */
    space_mode, 	/* skippen von leerzeichen                  */
    extra_space_mode, 	/* skippen von leerzeichen                  */
    other_mode, 	/* skippen von others(z.b. '(')          */
    skip_mode, 		/* skiped environemnt */
    comment_mode, 	/* behandeln von kommentaren                */
    active_mode, 	/* untersuchen von active-characters        */
    display_mode, 	
    math_mode, 		/* skippen der parameter von \end           */
    parameters_mode,
    in_optional_parameters_mode,
    in_default_parameters_mode
};

struct active_expansion
   {
    unsigned char active_char_parameter;
    unsigned char *active_expansion;
   };

struct active_definition
   {
    struct   active_definition *next;
    unsigned char              active_char;
             int               num_of_lines;
    struct   active_expansion  expansion[1];
   };
#define IS_DIGIT	4
#define IS_PUNC		8
#define IS_SENTENCE	16
#define IS_T_BEG	32
#define IS_T_END	64

Global void init_parser(FILE *sourcefile, FILE *outfile, 
							const char *sourcename);
Global void init_parser_data(void);

Global void read_parser_info(FILE *infofile);
Global void parse(struct token *end_token);

Global volatile int     line_number;
Global volatile int     column_number;
Global          Logical punctuation_check;

Global struct token catcodes[256];

#define WORD_MODE	0
#define NORMAL_MODE	1
#define EOF_MODE        2

#define RINGSIZE      15
#define MAXSTRING     256
#define MAX_WORD_LEN  1022
#define MODESTACK     100
#define MIN_WORD_LEN  1

#define DEF_LIB_LEN	600

Global Logical first_char_checking;
Global Logical case_sensitive;
Global library       *mainlib;
Global Logical newlibcreate;
Global library       *newlib;


#define Tolower(c) 	((catcodes[c].upper == (unsigned char)c) ? \
					catcodes[c].lower : (unsigned char)c)
#define Toupper(c) 	((catcodes[c].lower == (unsigned char)c) ? \
					catcodes[c].upper : (unsigned char)c)

